import os
import re
import ast
import json
import time
import random
import argparse
import statistics
import warnings
import subprocess as sp
import datetime

import numpy as np
import pandas as pd
import pyarrow as pa
import pyarrow.dataset as ds

import torch
import torch.nn as nn
from torch.optim import Adam
from torch.utils.data import DataLoader

from sklearn.model_selection import train_test_split

from tqdm import tqdm, tqdm_pandas
from tqdm.auto import tqdm

from transformers import (
    PreTrainedTokenizer, T5Tokenizer, T5EncoderModel, T5ForConditionalGeneration,
    BertModel, AutoTokenizer, AutoModel, GPT2Tokenizer,
    TrainingArguments, Trainer, get_scheduler,
    AutoModelForCausalLM, AutoConfig, AutoModelForSequenceClassification
)
import datasets
from datasets import load_metric

warnings.filterwarnings(
    "ignore", 
    message=(
        "The sentencepiece tokenizer that you are converting to a fast tokenizer "
        "uses the byte fallback option which is not implemented in the fast tokenizers."
    )
)
def combine_query_document(query: str, document: str, answer: str = None) -> str:
    """
    Combines a query and a document into a single string, optionally including an answer.

    Parameters:
    query (str): The query string.
    document (str): The document string.
    answer (str, optional): The answer string. Defaults to None.

    Returns:
    str: A combined string of the query, cleaned document, and optionally the answer.
    """
    # Clean the document by removing extra newlines, carriage returns, and tabs
    cleaned_document = re.sub(r'\n+', '\n', document.replace("\r", " ").replace("\t", " ")).strip()
    cleaned_document = cleaned_document.replace("=", " ").replace("-", " ")
    cleaned_document = re.sub(r'\s+', ' ', cleaned_document).strip()
    cleaned_document = " ".join(cleaned_document.split(" ")[:512])

    # Truncate the query if it is too long
    if len(query.split(" ")) > 100:
        query = " ".join(query.split(" ")[:30])

    # Combine query and cleaned document, optionally including the answer
    if answer is None:
        return query + " | " + cleaned_document
    else:
        try:
            return query + " | " + cleaned_document + " | " + answer
        except Exception as e:
            breakpoint()
            print("Error with combine_query_document")
            print("Query: " + str(query))
            print("Cleaned Document: " + str(cleaned_document))
            print("Answer: " + str(answer))
            return str(query) + " | " + str(cleaned_document) + " | " + str(answer)

def format_text_for_fine_tuning_content_relevance_sequence_classification(question: str, document: str) -> str:
    """
    Formats text for fine-tuning content relevance sequence classification.

    Parameters:
    question (str): The question string.
    document (str): The document string.

    Returns:
    str: A formatted string containing the instruction, question, and cleaned document.
    """
    instruction = (
        "You are an expert judge for evaluating question answering systems. "
        "Given the following question and document, you must analyze the provided document and determine whether it is sufficient for answering the question. \n\n"
    )

    # Clean the document by removing extra newlines, carriage returns, and tabs
    cleaned_document = re.sub(r'\n+', '\n', document.replace("\r", " ").replace("\t", " ")).strip()
    cleaned_document = cleaned_document.replace("=", " ").replace("-", " ")
    cleaned_document = re.sub(r'\s+', ' ', cleaned_document).strip()
    cleaned_document = " ".join(cleaned_document.split(" ")[:512])

    # Construct the instruction string
    instruction += "### Instruction:\n"
    instruction += f"Question: {question}\n"
    instruction += f"Document: {cleaned_document}\n"
    instruction += "### Response:\n"

    return instruction

def set_random_seed(seed: int):
    """
    Set the random seed for reproducibility.

    Parameters:
    seed (int): The seed value to set.
    """
    np.random.seed(seed)
    random.seed(seed)
    torch.manual_seed(seed)
    os.environ['PYTHONHASHSEED'] = str(seed)

def get_gpu_memory() -> list:
    """
    Retrieves the free GPU memory for each GPU available on the system.

    This function uses the `nvidia-smi` command to query the free memory of each GPU and returns a list of free memory values.

    Returns:
    list: A list of integers representing the free memory (in MiB) for each GPU.
    """
    command = "nvidia-smi --query-gpu=memory.free --format=csv"
    memory_free_info = sp.check_output(command.split()).decode('ascii').split('\n')[:-1][1:]
    memory_free_values = [int(x.split()[0]) for x in memory_free_info]
    return memory_free_values

class CustomBERTModel(nn.Module):
    def __init__(self, number_of_labels: int, model_choice: str):
        """
        Initializes the CustomBERTModel with the specified number of labels and model choice.

        Args:
            number_of_labels (int): The number of labels for the classification task.
            model_choice (str): The choice of the pre-trained model to use.
        """
        self.model_choice = model_choice
        super(CustomBERTModel, self).__init__()

        if model_choice in ["mosaicml/mpt-7b-instruct", "mosaicml/mpt-7b"]:
            config = AutoConfig.from_pretrained(model_choice, trust_remote_code=True)
            config.attn_config['attn_impl'] = 'triton'  # Use triton-based FlashAttention
            config.max_seq_len = max_token_length

            model_encoding = AutoModelForCausalLM.from_pretrained(
                model_choice,
                config=config,
                torch_dtype=torch.bfloat16,
                trust_remote_code=True,
                use_auth_token=True
            )
            embedding_size = 4096
            self.encoderModel = model_encoding.transformer

        elif model_choice in ['mosaicml/mpt-1b-redpajama-200b']:
            model_encoding = MptForSequenceClassification.from_pretrained(
                "mosaicml/mpt-1b-redpajama-200b", 
                trust_remote_code=True
            )
            embedding_size = 2048
            self.encoderModel = model_encoding

        elif model_choice in ["google/t5-large-lm-adapt", "google/t5-xl-lm-adapt"]:
            model_encoding = AutoModelForSequenceClassification.from_pretrained(model_choice)
            embedding_size = 1024
            self.encoderModel = model_encoding

        elif model_choice in ["roberta-large", "microsoft/deberta-v3-large"]:
            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 1024
            self.encoderModel = model_encoding

        elif model_choice in ["microsoft/deberta-v2-xlarge", "microsoft/deberta-v2-xxlarge"]:
            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 1536
            self.encoderModel = model_encoding

        else:
            model_encoding = AutoModel.from_pretrained(model_choice)
            embedding_size = 768
            self.encoderModel = model_encoding

        self.classifier = nn.Sequential(
            nn.Linear(embedding_size, 256),
            nn.Linear(256, number_of_labels)
        )
        self.embedding_size = embedding_size

    def forward(self, ids: torch.Tensor, mask: torch.Tensor, labels: torch.Tensor = None, 
    decoder_input_ids: torch.Tensor = None) -> torch.Tensor:
        """
        Perform a forward pass through the model.

        Parameters:
        ids (torch.Tensor): Input IDs tensor.
        mask (torch.Tensor): Attention mask tensor.
        labels (torch.Tensor, optional): Labels tensor. Defaults to None.
        decoder_input_ids (torch.Tensor, optional): Decoder input IDs tensor. Defaults to None.

        Returns:
        torch.Tensor: The output logits or classifier output.
        """
        # Retrieve the model choice
        model_choice = self.model_choice

        # Check if the model choice is one of the specified models
        if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt", "mosaicml/mpt-1b-redpajama-200b"]:
            # Perform a forward pass for the specified models
            total_output = self.encoderModel(input_ids=ids, attention_mask=mask)
            return total_output['logits']
        else:
            # Perform a forward pass for other models
            total_output = self.encoderModel(ids, attention_mask=mask)
            sequence_output = total_output['last_hidden_state']

            # Format the last hidden state and pass it through the classifier
            last_hidden_state_formatted = sequence_output[:, 0, :].view(-1, self.embedding_size)
            linear2_output = self.classifier(last_hidden_state_formatted)

            return linear2_output

def tokenize_function(tokenizer: AutoTokenizer, examples: dict) -> dict:
    """
    Tokenizes the input examples using the provided tokenizer.

    Parameters:
    tokenizer (AutoTokenizer): The tokenizer to use for tokenizing the examples.
    examples (dict): A dictionary containing the text to be tokenized. 
                     It should have a key "text" with the corresponding text data.

    Returns:
    dict: A dictionary containing the tokenized input with padding and truncation applied.
    """
    return tokenizer(examples["text"], padding="max_length", truncation=True)

def checkpoints(classification_datasets: list, model_choice: str) -> None:
    """
    Creates necessary checkpoint directories for the given classification datasets and model choice.

    Parameters:
    classification_datasets (list): A list of dataset names for which checkpoints need to be created.
    model_choice (str): The model choice string used to determine the folder structure.

    Returns:
    None
    """
    checkpoints_folder_path = "checkpoints/"
    
    # Create the main checkpoints directory if it doesn't exist
    if not os.path.isdir(checkpoints_folder_path):
        os.mkdir(checkpoints_folder_path)

    # Create a directory for the specific model choice
    dataset_folder_path = os.path.join(checkpoints_folder_path, model_choice.replace("/", "-"))
    if not os.path.isdir(dataset_folder_path):
        print(f"Creating folder: {dataset_folder_path}")
        os.mkdir(dataset_folder_path)

    # Create directories for each dataset within the model choice directory
    for dataset in classification_datasets:
        dataset_path = os.path.join(dataset_folder_path, dataset.replace("../", "").replace("/", "-"))
        try:
            os.mkdir(dataset_path)
        except FileExistsError:
            print("Already exists")
            print(dataset_path)

def load_model(model_choice: str) -> tuple[AutoTokenizer, int]:
    """
    Loads the tokenizer for the specified model choice and sets the maximum token length.

    Parameters:
    model_choice (str): The model identifier to load the tokenizer from.

    Returns:
    tuple: A tuple containing the tokenizer and the maximum token length.
           - tokenizer (AutoTokenizer): The tokenizer loaded from the specified model.
           - max_token_length (int): The maximum token length set for the tokenizer.
    """
    max_token_length = 2048
    tokenizer = AutoTokenizer.from_pretrained(model_choice, model_max_length=max_token_length)
    
    return tokenizer, max_token_length

def prepare_and_clean_data(params: dict) -> tuple[str, int]:
    """
    Prepares and cleans data based on the provided parameters.

    Parameters:
    params (dict): A dictionary containing the following keys:
        - "training_dataset_path" (str): Path to the training dataset.
        - "learning_rate_choices" (list): List of possible learning rates.
        - "chosen_learning_rate" (float): The chosen learning rate for the model.
        - "model_choice" (str): The model identifier.
        - "number_of_runs" (int): Number of runs for the training.
        - "validation_set_scoring" (str): Scoring method for the validation set.
        - "label" (str): The label column name in the dataset.
        - "validation_dataset_path" (str): Path to the validation dataset.
        - "patience_value" (int): Patience value for early stopping.
        - "num_epochs" (int): Number of epochs for training.
        - "num_warmup_steps" (int): Number of warmup steps for the learning rate scheduler.
        - "gradient_accumulation_multiplier" (int): Multiplier for gradient accumulation.
        - "assigned_batch_size" (int): Batch size assigned for training.
        - "tokenizer" (AutoTokenizer): Tokenizer to be used.

    Returns:
    tuple: A tuple containing:
        - checkpoint_path (str): Path to the checkpoint file.
        - patience_value (int): The patience value for early stopping.
    """
    # Extract parameters from the dictionary
    dataset = params["training_dataset_path"]
    learning_rate_choices = params["learning_rate_choices"]
    chosen_learning_rate = params["chosen_learning_rate"]
    model_choice = params["model_choice"]
    number_of_runs = params["number_of_runs"]
    validation_set_scoring = params["validation_set_scoring"]
    label_column = params["label"]
    validation_set = params["validation_dataset_path"]
    patience_value = params["patience_value"]
    num_epochs = params["num_epochs"]
    num_warmup_steps = params["num_warmup_steps"]
    gradient_accumulation_multiplier = params["gradient_accumulation_multiplier"]
    assigned_batch_size = params["assigned_batch_size"]
    tokenizer = params["tokenizer"]

    # Log the start of a new learning rate
    print("--------------------------------------------------------------------------")
    print("Starting new learning rate: " + str(chosen_learning_rate))
    print("--------------------------------------------------------------------------")

    # Generate current datetime for checkpoint naming
    current_datetime = datetime.datetime.now().strftime("%Y-%m-%d_%H:%M:%S")

    # Define parent directory for checkpoints
    parent_dir = "checkpoints/" + model_choice.replace("/", "-")

    # Create parent directory if it doesn't exist
    if not os.path.exists(parent_dir):
        print(f"Creating parent checkpoint directory: {parent_dir}")
        print("--------------------------------------------------------------------------")
        os.makedirs(parent_dir)

    # Define the checkpoint path
    checkpoint_path = os.path.join(
        "checkpoints",
        model_choice.replace("/", "-"),
        f"{label_column}_{os.path.basename(validation_set).replace('.tsv', '')}_{current_datetime}.pt"
    )

    # Record the start time of execution
    execution_start = time.time()

    # Log various parameters
    print("Dataset: " + dataset)
    print("Model: " + model_choice)
    print("Test Set Selection: " + validation_set)
    print("Number of Runs: " + str(number_of_runs))
    print('Learning Rate: ' + str(chosen_learning_rate))
    print("Checkpoint Path: " + checkpoint_path)
    print("Patience: " + str(patience_value))
    print("Validation Set Choice: " + str(validation_set_scoring))
    print("Number of Epochs: " + str(num_epochs))
    print("Number of warmup steps: " + str(num_warmup_steps))
    print("--------------------------------------------------------------------------")

    return checkpoint_path, patience_value

def analyze_and_report_data(dataset: str, label_column: str, tokenizer: AutoTokenizer, max_token_length: int) -> pd.DataFrame:
    """
    Analyzes and reports data from a given dataset.

    Parameters:
    - dataset (str): Path to the dataset file.
    - label_column (str): The column name in the dataset that contains the labels.
    - tokenizer: Tokenizer object used for encoding text.
    - max_token_length (int): Maximum token length allowed for the text.

    Returns:
    - pd.DataFrame: A DataFrame containing the processed and filtered data.
    """
    # Read the dataset
    synth_queries = pd.read_csv(dataset, sep="\t")

    # If the dataset is reformatted, rename columns accordingly
    if "nq_reformatted" in dataset:
        synth_queries['synthetic_query'] = synth_queries['Query']
        synth_queries['generated_answer'] = synth_queries['Answer']
        synth_queries['document'] = synth_queries['Document']

    # Print initial count
    print(f"Initial count: {len(synth_queries)}")

    # Filter out rows with NaN values in the specified columns
    synth_queries = synth_queries[synth_queries[label_column] != "NaN"]
    synth_queries = synth_queries[synth_queries["synthetic_query"].notna()]
    synth_queries = synth_queries[synth_queries["document"].notna()]
    synth_queries = synth_queries[synth_queries['generated_answer'].notna()]
    synth_queries = synth_queries[synth_queries[label_column].notna()]

    # Print count after initial filtering
    print(f"Count after initial filtering: {len(synth_queries)}")

    # Shuffle the dataset
    synth_queries = synth_queries.sample(n=len(synth_queries), random_state=42)

    # Print counts of Answer_Relevance_Label before any further filtering
    print(f"Answer_Relevance_Label counts before filtering: Yes - {synth_queries[synth_queries['Answer_Relevance_Label'] == 'Yes'].shape[0]}, No - {synth_queries[synth_queries['Answer_Relevance_Label'] == 'No'].shape[0]}")

    # Combine query and document (and generated answer if applicable) into a single text field
    if "Context" in label_column:
        synth_queries["concat_text"] = [
            combine_query_document(synth_queries.iloc[i]['synthetic_query'], synth_queries.iloc[i]['document'])
            for i in range(len(synth_queries))
        ]

        # Print the count before filtering duplicates
        print(f"Count before filtering duplicates for context relevance: {len(synth_queries)}")

        # Temporarily remove rows with duplicate query/document pairs for context relevance
        synth_queries = synth_queries.drop_duplicates(subset=["synthetic_query", "document"])

        # Print the count after filtering
        print(f"Count after filtering duplicates for context relevance: {len(synth_queries)}")

    else:
        synth_queries["concat_text"] = [
            combine_query_document(synth_queries.iloc[i]['synthetic_query'], synth_queries.iloc[i]['document'], synth_queries.iloc[i]['generated_answer'])
            for i in range(len(synth_queries))
        ]

        # Print the count before filtering
        print(f"Count before filtering for context relevance: {len(synth_queries)}")

        # Print counts of Context_Relevance_Label before filtering
        print(f"Context_Relevance_Label counts before filtering: Yes - {synth_queries[synth_queries['Context_Relevance_Label'] == 'Yes'].shape[0]}, No - {synth_queries[synth_queries['Context_Relevance_Label'] == 'No'].shape[0]}")

        # Temporarily remove rows where context relevance is "No" for answer relevance/faithfulness
        synth_queries = synth_queries[synth_queries["Context_Relevance_Label"] != "No"]

        # Print the count after filtering
        print(f"Count after filtering for context relevance: {len(synth_queries)}")

    # Print counts of Answer_Relevance_Label after filtering for context relevance
    print(f"Answer_Relevance_Label counts after filtering: Yes - {synth_queries[synth_queries['Answer_Relevance_Label'] == 'Yes'].shape[0]}, No - {synth_queries[synth_queries['Answer_Relevance_Label'] == 'No'].shape[0]}")

    # Tokenize the concatenated text and calculate token lengths
    synth_queries['token_length'] = [
        len(tokenizer.encode(text, return_tensors='pt')[0])
        for text in tqdm(synth_queries['concat_text'], desc="Tokenizing")
    ]

    # Print count before token length filtering
    print(f"Count before token length filtering: {len(synth_queries)}")

    # Remove duplicate rows based on the concatenated text
    synth_queries = synth_queries.drop_duplicates(["concat_text"])

    # Filter out rows where token length exceeds the maximum allowed token length
    synth_queries = synth_queries[synth_queries['token_length'] <= 4096]

    # Print final count
    print(f"Final count after token length filtering: {len(synth_queries)}")

    return synth_queries

def transform_data(synth_queries: pd.DataFrame, validation_set: str, label_column: str) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Transforms the synthetic queries and validation set for training and testing.

    Parameters:
    - synth_queries (pd.DataFrame): DataFrame containing synthetic queries.
    - validation_set (str): Path to the validation set file.
    - label_column (str): The column name used for labeling.

    Returns:
    - tuple: A tuple containing the transformed training and test DataFrames.
    """
    # Initialize the training DataFrame
    train_df = synth_queries

    # Read and preprocess the validation set
    test_set = pd.read_csv(validation_set, sep="\t")
    test_set['Question'] = test_set['Query']
    test_set['Document'] = test_set['Document'].str.strip()
    test_set = test_set[test_set["Document"].str.len() > 100]
    test_set = test_set[test_set[label_column].notna()]

    # Preprocess the training DataFrame
    train_df['document'] = train_df['document'].astype(str).str.strip()
    train_df = train_df[train_df["document"].str.len() > 100]
    train_df = train_df[train_df[label_column].notna()]

    # Print counts of Answer_Relevance_Label before any further filtering
    print(f"Answer_Relevance_Label counts before filtering: Yes - {train_df[train_df['Answer_Relevance_Label'] == 'Yes'].shape[0]}, No - {train_df[train_df['Answer_Relevance_Label'] == 'No'].shape[0]}")

    # Combine query and document (and generated answer if applicable) into a single text field
    if "Context" in label_column:
        test_set['concat_text'] = [
            combine_query_document(test_set.iloc[i]['Question'], test_set.iloc[i]['Document'])
            for i in range(len(test_set))
        ]

        # Print the count before filtering duplicates
        print(f"Count before filtering duplicates for context relevance: {len(train_df)}")

        # Temporarily remove rows with duplicate query/document pairs for context relevance
        train_df = train_df.drop_duplicates(subset=["synthetic_query", "document"])

        # Print the count after filtering
        print(f"Count after filtering duplicates for context relevance: {len(train_df)}")

    else:
        test_set['concat_text'] = [
            combine_query_document(test_set.iloc[i]['Question'], test_set.iloc[i]['Document'], test_set.iloc[i]['Answer'])
            for i in range(len(test_set))
        ]

        # Print the count before filtering
        print(f"Count before filtering for context relevance: {len(train_df)}")

        # Temporarily remove rows where context relevance is 0 for answer relevance/faithfulness
        train_df = train_df[train_df["Context_Relevance_Label"] != "No"]

        # Print the count after filtering
        print(f"Count after filtering for context relevance: {len(train_df)}")

    # Print counts of Answer_Relevance_Label after filtering for context relevance
    print(f"Answer_Relevance_Label counts after filtering: Yes - {train_df[train_df['Answer_Relevance_Label'] == 'Yes'].shape[0]}, No - {train_df[train_df['Answer_Relevance_Label'] == 'No'].shape[0]}")

    # Remove duplicate rows based on the concatenated text
    train_df = train_df.drop_duplicates(["concat_text"])
    test_set = test_set.drop_duplicates(["concat_text"])

    # Additional filtering for Answer_Faithfulness classification
    if "Faith" in label_column:
        print("Refining data for Answer_Faithfulness classification!")
        train_df = train_df[train_df["Context_Relevance_Label"].notna()]
        train_df = train_df[train_df["Answer_Faithfulness_Label"].notna()]
        error_strings = ['answer', 'contrad', 'false', 'information', 'unanswer', 'Answer', 'Contrad', 'False', 'Information', 'Unanswer']
        train_df['generated_answer'] = train_df['generated_answer'].astype(str)
        train_df = train_df[~train_df['generated_answer'].str.contains('|'.join(error_strings))]

    return train_df, test_set
def split_dataset(train_df: pd.DataFrame, dataset: str, 
test_set: pd.DataFrame, label_column: str) -> tuple[list[str], list[int], list[str], list[int], list[str], list[int], list[int]]:
    """
    Splits the dataset into training, development, and test sets, and extracts the corresponding labels.

    Parameters:
    - train_df (pd.DataFrame): DataFrame containing the training data.
    - dataset (str): Name of the dataset.
    - test_set (pd.DataFrame): DataFrame containing the test data.
    - label_column (str): The column name used for labeling.

    Returns:
    - tuple: A tuple containing:
        - train_set_text (list): List of concatenated text fields for the training set.
        - train_set_label (list): List of labels for the training set.
        - dev_set_text (list): List of concatenated text fields for the development set.
        - dev_set_label (list): List of labels for the development set.
        - test_set_text (list): List of concatenated text fields for the test set.
        - test_set_label (list): List of labels for the test set.
        - labels_list (list): Sorted list of unique labels.
    """
    
    # Conversion dictionary for binary classification
    conversion_dict = {"Yes": 1, "No": 0}
    
    # Extract concatenated text fields for the training set
    train_set_text = [train_df.iloc[i]['concat_text'] for i in range(len(train_df))]
    
    # Extract labels for the training set
    if "nq_reformatted" not in dataset:
        train_set_label = [conversion_dict[train_df.iloc[i][label_column]] for i in range(len(train_df))]
    else:
        train_set_label = [int(train_df.iloc[i][label_column]) for i in range(len(train_df))]
    
    # Extract concatenated text fields and labels for the development set
    dev_set_text = [test_set.iloc[i]['concat_text'] for i in range(len(test_set))]
    dev_set_label = [int(test_set.iloc[i][label_column]) for i in range(len(test_set))]
    
    # Extract concatenated text fields and labels for the test set
    test_set_text = [test_set.iloc[i]['concat_text'] for i in range(len(test_set))]
    test_set_label = [int(test_set.iloc[i][label_column]) for i in range(len(test_set))]

    # Create a sorted list of unique labels
    labels_list = sorted(list(set(train_set_label + dev_set_label + test_set_label)))

    return train_set_text, train_set_label, dev_set_text, dev_set_label, test_set_text, test_set_label, labels_list

    ############################################################

def prepare_dataset(validation_set_scoring: bool, 
                    train_set_label: list[int], 
                    train_set_text: list[str], 
                    dev_set_label: list[int], 
                    dev_set_text: list[str], 
                    test_set_label: list[int] = None, 
                    test_set_text: list[str] = None) -> tuple[pd.DataFrame, datasets.Dataset, datasets.Dataset, datasets.Dataset, pd.DataFrame]:
    """
    Prepares the dataset for training, validation, and testing by converting them into pandas DataFrames and Arrow tables.

    Parameters:
    - validation_set_scoring (bool): Flag to determine if validation set scoring is enabled.
    - train_set_label (list[int]): List of labels for the training set.
    - train_set_text (list[str]): List of text data for the training set.
    - dev_set_label (list[int]): List of labels for the development set.
    - dev_set_text (list[str]): List of text data for the development set.
    - test_set_label (list[int], optional): List of labels for the test set. Required if validation_set_scoring is False.
    - test_set_text (list[str], optional): List of text data for the test set. Required if validation_set_scoring is False.

    Returns:
    - tuple: A tuple containing:
        - training_dataset_pandas (pd.DataFrame): DataFrame for the training set.
        - training_dataset_arrow (datasets.Dataset): Arrow table for the training set.
        - validation_dataset_arrow (datasets.Dataset): Arrow table for the validation set.
        - test_dataset_arrow (datasets.Dataset): Arrow table for the test set.
        - test_dataset_pandas (pd.DataFrame): DataFrame for the test set.
    """
    
    if validation_set_scoring:
        # Prepare training dataset
        training_dataset_pandas = pd.DataFrame({'label': train_set_label, 'text': train_set_text})
        training_dataset_arrow = datasets.Dataset(pa.Table.from_pandas(training_dataset_pandas))

        # Prepare validation dataset
        validation_dataset_pandas = pd.DataFrame({'label': dev_set_label, 'text': dev_set_text})
        validation_dataset_arrow = datasets.Dataset(pa.Table.from_pandas(validation_dataset_pandas))

        # Prepare test dataset
        test_dataset_pandas = pd.DataFrame({'label': dev_set_label, 'text': dev_set_text})
        test_dataset_arrow = datasets.Dataset(pa.Table.from_pandas(test_dataset_pandas))
    else:
        # Prepare training dataset
        training_dataset_pandas = pd.DataFrame({'label': train_set_label, 'text': train_set_text})
        training_dataset_arrow = datasets.Dataset(pa.Table.from_pandas(training_dataset_pandas))

        # Prepare validation dataset
        validation_dataset_pandas = pd.DataFrame({'label': dev_set_label, 'text': dev_set_text})
        validation_dataset_arrow = datasets.Dataset(pa.Table.from_pandas(validation_dataset_pandas))

        # Prepare test dataset
        test_dataset_pandas = pd.DataFrame({'label': test_set_label, 'text': test_set_text})
        test_dataset_arrow = datasets.Dataset(pa.Table.from_pandas(test_dataset_pandas))

    return training_dataset_pandas, training_dataset_arrow, validation_dataset_arrow, test_dataset_arrow, test_dataset_pandas

def initalize_dataset_for_tokenization(tokenizer: PreTrainedTokenizer, 
                                      training_dataset_arrow: datasets.Dataset, 
                                      validation_dataset_arrow: datasets.Dataset, 
                                      test_dataset_arrow: datasets.Dataset) -> datasets.DatasetDict:
    """
    Initializes and tokenizes the dataset for training, validation, and testing.

    Parameters:
    - tokenizer (PreTrainedTokenizer): The tokenizer to be used for tokenizing the text data.
    - training_dataset_arrow (datasets.Dataset): Arrow table for the training set.
    - validation_dataset_arrow (datasets.Dataset): Arrow table for the validation set.
    - test_dataset_arrow (datasets.Dataset): Arrow table for the test set.

    Returns:
    - datasets.DatasetDict: A dictionary containing the tokenized datasets for training, validation, and testing.
    """
    
    # Create a DatasetDict with the provided datasets
    classification_dataset = datasets.DatasetDict({
        'train': training_dataset_arrow, 
        'validation': validation_dataset_arrow, 
        'test': test_dataset_arrow
    })
    
    # Tokenize the datasets
    tokenized_datasets = classification_dataset.map(lambda examples: tokenize_function(tokenizer, examples), batched=True)

    # Remove the 'text' column as it is no longer needed after tokenization
    tokenized_datasets = tokenized_datasets.remove_columns(["text"])
    
    # Rename the 'label' column to 'labels' to match the expected input format for the model
    tokenized_datasets = tokenized_datasets.rename_column("label", "labels")
    
    # Set the format of the datasets to PyTorch tensors
    tokenized_datasets.set_format("torch")

    return tokenized_datasets

    ############################################################

def train_and_evaluate_model(params: dict) -> tuple[list[torch.nn.Module], list[float], list[float], list[float]]:
    """
    Trains and evaluates a model based on the provided parameters.

    Parameters:
    - params (dict): A dictionary containing the following keys:
        - number_of_runs (int): Number of training runs.
        - tokenized_datasets (datasets.DatasetDict): Tokenized datasets for training, validation, and testing.
        - assigned_batch_size (int): Batch size for training and evaluation.
        - train_set_label (list): Labels for the training set.
        - model_choice (str): Model choice identifier.
        - chosen_learning_rate (float): Learning rate for the optimizer.
        - device (torch.device): Device to run the model on (e.g., 'cpu' or 'cuda').
        - checkpoint_path (str): Path to save the model checkpoints.
        - patience_value (int): Patience value for early stopping.
        - num_epochs (int): Number of epochs for training.
        - num_warmup_steps (int): Number of warmup steps for the learning rate scheduler.
        - gradient_accumulation_multiplier (int): Gradient accumulation steps.

    Returns:
    - tuple: A tuple containing the trained model, average training losses, average validation losses, evaluation dataloader, and inference times.
    """
    
    # Extract parameters from the dictionary
    number_of_runs = params["number_of_runs"]
    tokenized_datasets = params["tokenized_datasets"]
    assigned_batch_size = params["assigned_batch_size"]
    train_set_label = params["train_set_label"]
    model_choice = params["model_choice"]
    chosen_learning_rate = params["chosen_learning_rate"]
    device = params["device"]
    checkpoint_path = params["checkpoint_path"]
    patience_value = params["patience_value"]
    num_epochs = params["num_epochs"]
    num_warmup_steps = params["num_warmup_steps"]
    gradient_accumulation_multiplier = params["gradient_accumulation_multiplier"]

    # Initialize lists to store metrics
    micro_averages = []
    macro_averages = []
    inference_times = []

    for i in range(number_of_runs):
        run_start = time.time()
        print("Loading Model")

        # Create dataloaders for training, validation, and evaluation
        train_dataloader = DataLoader(tokenized_datasets['train'], batch_size=assigned_batch_size)
        validation_dataloader = DataLoader(tokenized_datasets['validation'], batch_size=assigned_batch_size)
        eval_dataloader = DataLoader(tokenized_datasets['test'], batch_size=assigned_batch_size)

        # Initialize the model and move it to the specified device
        model = CustomBERTModel(len(set(train_set_label)), model_choice)
        model.to(device)

        # Define the loss function and optimizer
        criterion = nn.CrossEntropyLoss()
        optimizer = Adam(model.parameters(), lr=chosen_learning_rate)

        # Calculate the total number of training steps and initialize the learning rate scheduler
        num_training_steps = num_epochs * len(train_dataloader)
        lr_scheduler = get_scheduler(
            name="linear", optimizer=optimizer, num_warmup_steps=num_warmup_steps, num_training_steps=num_training_steps
        )

        # Initialize lists to store losses
        train_losses = []
        valid_losses = []
        avg_train_losses = []
        avg_valid_losses = []

        # Import EarlyStopping utility
        from ares.LLM_as_a_Judge_Adaptation.pytorchtools import EarlyStopping

        # Initialize early stopping
        early_stopping = EarlyStopping(patience=patience_value, verbose=True, path=checkpoint_path)
        print("Checkpoint Path: " + checkpoint_path)
        print("Beginning Training")

        total_epochs_performed = 0

        for epoch in range(num_epochs):
            total_epochs_performed += 1
            print("Current Epoch: " + str(epoch))

            progress_bar = tqdm(range(len(train_dataloader)))
            gradient_accumulation_count = 0
            model.train()
            for batch in train_dataloader:
                # Prepare the batch for the model
                if model_choice in ["mosaicml/mpt-1b-redpajama-200b"]:
                    new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].bool().to(device)}
                else:
                    new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].to(device)}

                # Forward pass
                outputs = model(**new_batch)
                loss = criterion(outputs, batch['labels'].to(device))
                loss.backward()

                # Gradient accumulation
                gradient_accumulation_count += 1
                if gradient_accumulation_count % gradient_accumulation_multiplier == 0:
                    optimizer.step()
                    lr_scheduler.step()
                    optimizer.zero_grad()
                    
                progress_bar.update(1)
                train_losses.append(loss.item())

            progress_bar = tqdm(range(len(validation_dataloader)))
            model.eval()
            for batch in validation_dataloader:
                with torch.no_grad():
                    # Prepare the batch for the model
                    if model_choice in ["mosaicml/mpt-1b-redpajama-200b"]:
                        new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].bool().to(device)}
                    else:
                        new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].to(device)}

                    if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt"]:
                        new_batch['decoder_input_ids'] = batch['labels'].reshape(batch['labels'].shape[0], 1).to(device)

                    # Forward pass
                    outputs = model(**new_batch)
                    loss = criterion(outputs, batch['labels'].to(device))
                    progress_bar.update(1)
                    valid_losses.append(loss.item())

            # Calculate average losses
            train_loss = np.average(train_losses)
            valid_loss = np.average(valid_losses)
            avg_train_losses.append(train_loss)
            avg_valid_losses.append(valid_loss)

            # Print epoch summary
            epoch_len = len(str(num_epochs))
            print_msg = (f'[{epoch:>{epoch_len}}/{num_epochs:>{epoch_len}}] ' +
                         f'train_loss: {train_loss:.5f} ' +
                         f'valid_loss: {valid_loss:.5f}')
            print(print_msg)

            # Reset losses for the next epoch
            train_losses = []
            valid_losses = []
            early_stopping(valid_loss, model)
            if early_stopping.early_stop:
                print("Early stopping")
                break

    return model, avg_train_losses, avg_valid_losses, eval_dataloader, inference_times

        ############################################################

def evaluate_model(model: torch.nn.Module, model_choice: str, checkpoint_path: str, 
device: torch.device, eval_dataloader: DataLoader, inference_times: list) -> tuple:
    """
    Evaluates the given model on the evaluation dataset.

    Parameters:
    - model (torch.nn.Module): The model to be evaluated.
    - model_choice (str): The choice of model architecture.
    - checkpoint_path (str): Path to the model checkpoint to load.
    - device (torch.device): The device to run the evaluation on (e.g., 'cuda:0').
    - eval_dataloader (DataLoader): DataLoader for the evaluation dataset.
    - inference_times (list): List to store inference times.

    Returns:
    - tuple: A tuple containing total predictions, total references, and the evaluation metric.
    """

    print("Loading the Best Model")
    model.load_state_dict(torch.load(checkpoint_path))

    print("Beginning Evaluation")
    metric = load_metric("accuracy")

    total_predictions = torch.FloatTensor([]).to(device)
    total_references = torch.FloatTensor([]).to(device)

    inference_start = time.time()

    progress_bar = tqdm(range(len(eval_dataloader)))
    for batch in eval_dataloader:
        with torch.no_grad():
            # Prepare the batch for the model
            if model_choice in ["mosaicml/mpt-1b-redpajama-200b"]:
                new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].bool().to(device)}
            else:
                new_batch = {'ids': batch['input_ids'].to(device), 'mask': batch['attention_mask'].to(device)}

            if model_choice in ["t5-small", "google/t5-xl-lm-adapt", "google/t5-large-lm-adapt"]:
                new_batch['decoder_input_ids'] = batch['labels'].reshape(batch['labels'].shape[0], 1).to(device)

            # Forward pass
            outputs = model(**new_batch)

            logits = outputs
            predictions = torch.argmax(logits, dim=-1)
            metric.add_batch(predictions=predictions, references=batch['labels'].to(device))

            total_predictions = torch.cat((total_predictions, predictions), 0)
            total_references = torch.cat((total_references, batch['labels'].to(device)), 0)

            progress_bar.update(1)

    inference_end = time.time()
    total_inference_time = inference_end - inference_start
    inference_times.append(total_inference_time)

    return total_predictions, total_references, metric

        ############################################################

def print_and_save_model(total_predictions: torch.Tensor, total_references: torch.Tensor, checkpoint_path: str, metric) -> None:
    """
    Prints the shapes of the predictions and references, computes and prints the accuracy and F1 scores,
    and saves the classification checkpoint.

    Args:
    - total_predictions (torch.Tensor): Tensor containing the model's predictions.
    - total_references (torch.Tensor): Tensor containing the true labels.
    - checkpoint_path (str): Path to save the classification checkpoint.
    - metric: Metric object used to compute the accuracy.

    Returns:
    - None
    """
    print("--------------------------")
    print("Predictions and Reference Shapes")
    print(total_predictions.shape)
    print(total_references.shape)

    # Compute and print accuracy
    results = metric.compute(references=total_references, predictions=total_predictions)
    print("Accuracy for Test Set: " + str(results['accuracy']))

    # Compute and print F1 scores
    f_1_metric = load_metric("f1", trust_remote_code=True)
    macro_f_1_results = f_1_metric.compute(average='macro', references=total_references, predictions=total_predictions)
    print("Macro F1 for Test Set: " + str(macro_f_1_results['f1'] * 100))
    micro_f_1_results = f_1_metric.compute(average='micro', references=total_references, predictions=total_predictions)
    print("Micro F1 for Test Set: " + str(micro_f_1_results['f1'] * 100))

    # Compute and print positive/negative reference ratio
    positive_ratio = round(total_references.tolist().count(1) / len(total_references.tolist()), 3)
    print("Positive / Negative Reference Ratio: " + str(positive_ratio))

    # Print checkpoint save path
    print("Saved classification checkpoint to: " + str(checkpoint_path))



import pandas as pd
import torch
import openai
import sys
from transformers import AutoTokenizer, AutoModelForCausalLM
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring 
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_togetherai
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring_togetherai
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring_togetherai
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_claude 
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring_claude
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring_claude
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_context_relevance_scoring_vllm
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_faithfulness_scoring_vllm
from ares.RAG_Automatic_Evaluation.Evaluation_Functions import few_shot_answer_relevance_scoring_vllm
failed_extraction_count = {'failed': 0}

if 'ipykernel' in sys.modules:
    # We are in a Jupyter notebook or similar (uses IPython kernel)
    from tqdm.notebook import tqdm
else:
    # We are in a regular Python environment (e.g., terminal or script)
    from tqdm import tqdm

def validate_inputs(vllm: bool, host_url: str, in_domain_prompts_dataset: str, unlabeled_evaluation_set: str, documents: int) -> tuple:
    """
    Validates the input parameters for the evaluation process.

    Parameters:
    vllm (bool): Indicates if vLLM is being used.
    host_url (str): The URL of the host if vLLM is used.
    in_domain_prompts_dataset (str): Path to the in-domain prompts dataset file.
    unlabeled_evaluation_set (str): Path to the unlabeled evaluation set file.
    documents (int): Number of documents to be processed.

    Returns:
    tuple: A tuple containing the in-domain prompts dataset and the unlabeled evaluation set as pandas DataFrames.

    Raises:
    ValueError: If the input parameters are not valid.
    SystemExit: If the documents size is larger than the documents present in the unlabeled evaluation set.
    """

    # Validate vLLM and host_url relationship
    if vllm and host_url == "None":
        raise ValueError("host_url must be provided if vllm is True.")
    elif not vllm and host_url != "None":
        raise ValueError("vLLM must be set to True if host_url is provided")

    # Load in-domain prompts dataset if provided
    if in_domain_prompts_dataset is not None:
        in_domain_prompts_dataset = pd.read_csv(in_domain_prompts_dataset, sep='\t')
    else:
        in_domain_prompts_dataset = None

    # Load unlabeled evaluation set if provided
    if unlabeled_evaluation_set is not None:
        unlabeled_evaluation_set = pd.read_csv(unlabeled_evaluation_set, sep='\t')
    else:
        unlabeled_evaluation_set = None

    # Check if both datasets are provided
    if in_domain_prompts_dataset is None and unlabeled_evaluation_set is None:
        print("Error: UES and IDP are not provided")
        exit()
    if in_domain_prompts_dataset is None:
        print("Error: IDP is not provided")
        exit()
    if unlabeled_evaluation_set is None:
        print("Error: UES is not provided")
        exit()

    # Set documents to the length of the unlabeled evaluation set if documents is 0
    if documents == 0 or documents == "None":
        documents = len(unlabeled_evaluation_set)

    # Validate documents size
    if documents > len(unlabeled_evaluation_set):
        sys.exit("Error: documents size passed in is larger than documents present in unlabeled evaluation set")

    return in_domain_prompts_dataset, unlabeled_evaluation_set, documents

def extract_query(row: pd.Series, in_domain_prompts_dataset: pd.DataFrame) -> tuple:
    """
    Extracts the query from a given row and determines the query identifier.

    Parameters:
    row (pd.Series): A row from the dataset containing the query or question.
    in_domain_prompts_dataset (pd.DataFrame): The in-domain prompts dataset to check for the query identifier.

    Returns:
    tuple: A tuple containing the extracted query and the query identifier.

    Raises:
    SystemExit: If both 'Query' and 'Question' keys are missing in the given row.
    """
    try:
        query = row['Query']
    except KeyError:
        query = row['Question']
    
    try:
        _ = in_domain_prompts_dataset.iloc[0]['Query']
        query_id = "Query"
    except KeyError:
        try:
            _ = in_domain_prompts_dataset.iloc[0]['Question']
            query_id = "Question"
        except KeyError:
            sys.exit("Both 'Query' and 'Question' keys are missing for the given row.")
    
    return query, query_id

def score_row(row: pd.Series, 
              in_domain_prompts_dataset: pd.DataFrame, 
              context_relevance_system_prompt: str, 
              answer_relevance_system_prompt: str, 
              answer_faithfulness_system_prompt: str, 
              model_choice: str, 
              query_id: str, 
              debug_mode: bool, 
              request_delay: int, 
              vllm: bool, 
              host_url: str) -> tuple:
    """
    Scores a row based on context relevance, answer relevance, and answer faithfulness.

    Parameters:
    row (pd.Series): A row from the dataset containing the query, document, and answer.
    in_domain_prompts_dataset (pd.DataFrame): The in-domain prompts dataset.
    context_relevance_system_prompt (str): The system prompt for context relevance scoring.
    answer_relevance_system_prompt (str): The system prompt for answer relevance scoring.
    answer_faithfulness_system_prompt (str): The system prompt for answer faithfulness scoring.
    model_choice (str): The model choice for scoring (e.g., "gpt", "claude").
    query_id (str): The query identifier.
    debug_mode (bool): Flag to enable or disable debug mode.
    request_delay (int): The delay between requests.
    vllm (bool): Flag to indicate if vllm is used.
    host_url (str): The host URL for the scoring service.

    Returns:
    tuple: A tuple containing context score, answer relevance score, and answer faithfulness score.
    """
    query, query_id = extract_query(row, in_domain_prompts_dataset)
    document = row['Document']
    answer = row['Answer']

    if vllm:
        context_score = few_shot_context_relevance_scoring_vllm(
            context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, host_url, request_delay, failed_extraction_count, in_domain_prompts_dataset)
        
        if context_score == 0:
            return 0, 0, 0
        else:
            answer_relevance_score = few_shot_answer_relevance_scoring_vllm(
                answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, host_url, request_delay, failed_extraction_count, in_domain_prompts_dataset)
            answer_faithfulness_score = few_shot_answer_faithfulness_scoring_vllm(
                answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, host_url, request_delay, failed_extraction_count, in_domain_prompts_dataset)
    else:
        if "gpt" in model_choice:
            context_score = few_shot_context_relevance_scoring(
                context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
        elif "claude" in model_choice:
            context_score = few_shot_context_relevance_scoring_claude(
                context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
        else:
            context_score = few_shot_context_relevance_scoring_togetherai(
                context_relevance_system_prompt, query, document, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
        
        if context_score == 0:
            return 0, 0, 0
        else:
            if "gpt" in model_choice:
                answer_relevance_score = few_shot_answer_relevance_scoring(
                    answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
                answer_faithfulness_score = few_shot_answer_faithfulness_scoring(
                    answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
            elif "claude" in model_choice:
                answer_relevance_score = few_shot_answer_relevance_scoring_claude(
                    answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
                answer_faithfulness_score = few_shot_answer_faithfulness_scoring_claude(
                    answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
            else:
                answer_relevance_score = few_shot_answer_relevance_scoring_togetherai(
                    answer_relevance_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
                answer_faithfulness_score = few_shot_answer_faithfulness_scoring_togetherai(
                    answer_faithfulness_system_prompt, query, document, answer, model_choice, query_id, debug_mode, request_delay, failed_extraction_count, in_domain_prompts_dataset)
    
    return context_score, answer_relevance_score, answer_faithfulness_score

def evaluate_documents(
    unlabeled_evaluation_set: pd.DataFrame, 
    in_domain_prompts_dataset: str, 
    context_relevance_system_prompt: str, 
    answer_relevance_system_prompt: str, 
    answer_faithfulness_system_prompt: str, 
    model_choice: str, 
    debug_mode: bool, 
    request_delay: int, 
    vllm: bool, 
    host_url: str, 
    documents: int
) -> tuple[list[float], list[float], list[float]]:
    """
    Evaluates a subset of documents for context relevance, answer relevance, and answer faithfulness.

    Parameters:
    unlabeled_evaluation_set (pd.DataFrame): The dataset containing the documents to be evaluated.
    in_domain_prompts_dataset (str): Path to the in-domain prompts dataset.
    context_relevance_system_prompt (str): System prompt for context relevance scoring.
    answer_relevance_system_prompt (str): System prompt for answer relevance scoring.
    answer_faithfulness_system_prompt (str): System prompt for answer faithfulness scoring.
    model_choice (str): The model choice for evaluation.
    debug_mode (bool): Flag to enable or disable debug mode.
    request_delay (int): Delay between requests in seconds.
    vllm (bool): Flag to indicate if vllm is used.
    host_url (str): The host URL for the model.
    documents (int): Number of documents to evaluate.

    Returns:
    tuple: A tuple containing three lists - context relevance scores, answer relevance scores, and answer faithfulness scores.
    """
    context_relevance_scores = []
    answer_relevance_scores = []
    answer_faithfulness_scores = []

    with tqdm(total=documents, desc=f"Evaluating large subset with {model_choice}") as pbar:
        for index, row in unlabeled_evaluation_set[:documents].iterrows():
            context_score, answer_relevance_score, answer_faithfulness_score = score_row(
                row, in_domain_prompts_dataset, context_relevance_system_prompt, answer_relevance_system_prompt, answer_faithfulness_system_prompt, model_choice, "Query", debug_mode, request_delay, vllm, host_url)
            
            context_relevance_scores.append(context_score)
            answer_relevance_scores.append(answer_relevance_score)
            answer_faithfulness_scores.append(answer_faithfulness_score)

            pbar.update(1)
    
    return context_relevance_scores, answer_relevance_scores, answer_faithfulness_scores

def ues_idp_config(
    in_domain_prompts_dataset: str, 
    unlabeled_evaluation_set: str, 
    context_relevance_system_prompt: str, 
    answer_relevance_system_prompt: str, 
    answer_faithfulness_system_prompt: str, 
    debug_mode: bool, 
    documents: int, 
    model_choice: str, 
    vllm: bool, 
    host_url: str, 
    request_delay: int
) -> dict:
    """
    Configures UES and IDP for evaluation.

    Parameters:
    in_domain_prompts_dataset (str): Path to the in-domain prompts dataset.
    unlabeled_evaluation_set (str): Path to the unlabeled evaluation set.
    context_relevance_system_prompt (str): System prompt for context relevance scoring.
    answer_relevance_system_prompt (str): System prompt for answer relevance scoring.
    answer_faithfulness_system_prompt (str): System prompt for answer faithfulness scoring.
    debug_mode (bool): Flag to enable or disable debug mode.
    documents (int): Number of documents to evaluate.
    model_choice (str): The model choice for evaluation.
    vllm (bool): Flag to indicate if vllm is used.
    host_url (str): The host URL for the model.
    request_delay (int): Delay between requests in seconds.

    Returns:
    dict: A dictionary containing the mean scores for context relevance, answer faithfulness, and answer relevance.
    """
    
    # Validate inputs and get the processed datasets
    in_domain_prompts_dataset, unlabeled_evaluation_set, documents = validate_inputs(
        vllm, host_url, in_domain_prompts_dataset, unlabeled_evaluation_set, documents
    )

    # Evaluate the documents and get the scores
    context_relevance_scores, answer_relevance_scores, answer_faithfulness_scores = evaluate_documents(
        unlabeled_evaluation_set, in_domain_prompts_dataset, context_relevance_system_prompt, 
        answer_relevance_system_prompt, answer_faithfulness_system_prompt, model_choice, 
        debug_mode, request_delay, vllm, host_url, documents
    )

    # Create a DataFrame to store the results
    results_df = pd.DataFrame({
        'Context_Relevance_Score': context_relevance_scores,
        'Answer_Relevance_Score': answer_relevance_scores,
        'Answer_Faithfulness_Score': answer_faithfulness_scores
    })

    # Filter out invalid results
    valid_results_df = results_df[
        ~((results_df['Context_Relevance_Score'] == 0) & 
        ((results_df['Answer_Relevance_Score'] == 1) | 
        (results_df['Answer_Faithfulness_Score'] == 1)))
    ]
    
    # Print the number of failed extractions
    print("Number of times did not extract Yes or No:", failed_extraction_count['failed'])

    # Return the mean scores rounded to three decimal places
    return {
        "Context Relevance Scores": round(valid_results_df['Context_Relevance_Score'].mean(), 3),
        "Answer Faithfulness Scores": round(valid_results_df['Answer_Faithfulness_Score'].mean(), 3),
        "Answer Relevance Scores": round(valid_results_df['Answer_Relevance_Score'].mean(), 3)
    }


import ast
import copy
import csv
import json
import math
import random
import re
import requests
import time

import numpy as np
import pandas as pd
import torch
from datasets import Dataset
from sklearn.model_selection import train_test_split
from tqdm import tqdm

import openai
from transformers import (AutoConfig, AutoModelForCausalLM, AutoModelForSeq2SeqLM,
                          AutoTokenizer, BitsAndBytesConfig)

from ares.LLM_as_a_Judge_Adaptation.LLM_Synthetic_Generation import generate_synthetic_contradictory_answers_api_approach

def generate_synthetic_query_llm_approach(document: str, prompt: str, length_of_fewshot_prompt: int, 
device: torch.device, tokenizer: AutoTokenizer, model: AutoModelForCausalLM, percentiles: list, 
for_fever_dataset=False, for_wow_dataset=False) -> list:
    """
    Generates synthetic queries based on a document using a language model.

    This function constructs a prompt with the document and generates queries based on the specified dataset type.
    It handles token length limitations by truncating the document if necessary.

    Args:
        document (str): The document to base the query on.
        prompt (str): The initial prompt to the language model.
        length_of_fewshot_prompt (int): The number of few-shot examples already included in the prompt.
        device: The device to run the model on (e.g., 'cuda' or 'cpu').
        tokenizer: The tokenizer for encoding the text.
        model: The language model used for generating text.
        percentiles (list): A list of percentiles for sampling during generation.
        for_fever_dataset (bool, optional): Flag to indicate if the function is used for the FEVER dataset. Defaults to False.
        for_wow_dataset (bool, optional): Flag to indicate if the function is used for the WoW dataset. Defaults to False.

    Returns:
        list: A list of synthetic queries generated by the model.
    """

    synthetic_queries = []

    # Construct the prompt without the document based on the dataset type
    prompt_without_document = prompt + "Example " + str(length_of_fewshot_prompt + 1) + ":\n"
    if for_fever_dataset:
        prompt_without_document += "Document: \nStatement: "
    elif for_wow_dataset:
        prompt_without_document += "Document: \nDialogue: "
    else:
        prompt_without_document += "Document: \nQuestion: "

    # Calculate the length of tokens for the prompt and document
    prompt_tokens_length = tokenizer.encode(prompt_without_document, return_tensors='pt').to(device).shape[1]
    document_length = tokenizer.encode(document, return_tensors='pt').to(device).shape[1]

    # Check if the total length exceeds the model's maximum input size and truncate if necessary
    if prompt_tokens_length + document_length + 100 >= 2048:
        encoded_input = tokenizer(document, max_length=2048 - prompt_tokens_length - 100, truncation=True, return_tensors='pt')
        truncated_document = tokenizer.decode(encoded_input['input_ids'][0][:2048 - prompt_tokens_length - 100]) 
        document = truncated_document.replace("</s>", "")

    # Append the document to the prompt
    prompt += "Example " + str(length_of_fewshot_prompt + 1) + ":\n"
    prompt += "Document: " + document + "\n"
    if for_fever_dataset:
        prompt += "Statement: "
    elif for_wow_dataset:
        prompt += "Dialogue: "
    else:
        prompt += "Question: "

    # Encode the complete prompt
    input_ids = tokenizer.encode(prompt, max_length=2048, truncation=True, return_tensors='pt').to(device)

    # Set the maximum length for the generated text based on the dataset
    max_length = 32 if not for_wow_dataset else 256

    # Generate queries for each percentile
    for percentile in percentiles:
        if input_ids.shape[0] != 1 or input_ids.shape[1] >= 2048:
            print("Length of problematic input ids: " + str(input_ids.shape))
            print("Length of problematic document: " + str(len(encoded_input['input_ids'][0])))
            assert False
        outputs = model.generate(
            input_ids=input_ids,
            max_length=max_length,
            do_sample=True,
            top_p=percentile,
            num_return_sequences=1)

        query = tokenizer.decode(outputs[0], skip_special_tokens=True)

        synthetic_queries.append(query)

    return synthetic_queries

def generate_answer_llm_approach(document: str, question: str, prompt: str, length_of_fewshot_prompt: int, 
device: torch.device, tokenizer: AutoTokenizer, model: AutoModelForCausalLM, for_fever_dataset=False, for_wow_dataset=False) -> str:
    """
    Generates an answer using a language model based on the provided document and question.

    This function constructs a prompt for the language model by appending the document and question to a base prompt.
    It then encodes the prompt and checks if the total token length exceeds the model's maximum input size.
    If it does, the document is truncated to fit. The function finally generates an answer using the model.

    Args:
        document (str): The document text to be used for generating the answer.
        question (str): The question text based on the document.
        prompt (str): The initial prompt text to which the document and question will be appended.
        length_of_fewshot_prompt (int): The ordinal number of the current example in the context of few-shot learning.
        device: str: The device (CPU/GPU) on which the tokenizer and model are loaded.
        tokenizer: The tokenizer used for encoding the text.
        model: The model used for generating the answer.
        for_fever_dataset (bool, optional): Flag to indicate if the function is being used for the FEVER dataset. Defaults to False.
        for_wow_dataset (bool, optional): Flag to indicate if the function is being used for the WoW dataset. Defaults to False.

    Returns:
        str: The generated answer text.
    """
    # Construct the prompt without the document based on the dataset type
    prompt_without_document = prompt + "Example " + str(length_of_fewshot_prompt + 1) + ":\n"
    if for_fever_dataset:
        prompt_without_document += "Document: \nStatement: \nAnswer: "
    elif for_wow_dataset:
        prompt_without_document += "Document: \nDialogue: \nResponse: "
    else:
        prompt_without_document += "Document: \nQuestion: \nAnswer: "

    # Calculate the token lengths for the prompt, document, and question
    prompt_tokens_length = tokenizer.encode(prompt_without_document, return_tensors='pt').to(device).shape[1]
    document_length = tokenizer.encode(document, return_tensors='pt').to(device).shape[1]
    question_length = tokenizer.encode(question, return_tensors='pt').to(device).shape[1]

    # Check if the total length exceeds the model's maximum input size and truncate if necessary
    if prompt_tokens_length + document_length + question_length + 100 >= 2048:
        reduction_length = prompt_tokens_length + question_length + 100
        encoded_input = tokenizer(document, max_length=2048 - reduction_length, truncation=True, return_tensors='pt')
        truncated_document = tokenizer.decode(encoded_input['input_ids'][0][:2048 - reduction_length]) 
        document = truncated_document.replace("</s>", "")

    # Append the document and question to the prompt
    prompt += "Example " + str(length_of_fewshot_prompt + 1) + ":\n"
    prompt += "Document: " + document + "\n"
    if for_fever_dataset:
        prompt += "Statement: " + question + "\n"
        prompt += "Answer: "
    elif for_wow_dataset:
        prompt += "Dialogue: " + question + "\n"
        prompt += "Response: "
    else:
        prompt += "Question: " + question + "\n"
        prompt += "Answer: "

    # Encode the complete prompt
    input_ids = tokenizer.encode(prompt, max_length=2048, truncation=True, return_tensors='pt').to(device)

    # Check for encoding issues and generate the answer
    if input_ids.shape[0] != 1 or input_ids.shape[1] >= 2048:
        print("Length of problematic input ids: " + str(input_ids.shape))
        print("Length of problematic document: " + str(len(encoded_input['input_ids'][0])))
        assert False
    outputs = model.generate(
        input_ids=input_ids,
        max_length=256,
        do_sample=True,
        top_p=0.05,
        num_return_sequences=1)

    answer = tokenizer.decode(outputs[0], skip_special_tokens=True)

    return answer

def generate_synthetic_query_openai_approach(document: str, system_prompt: str, 
few_shot_examples: str, temperatures: list, length_of_fewshot_prompt: int) -> list:
    """
    Generates synthetic queries using the OpenAI API with different temperatures.

    This function takes a document and a system prompt, combines them with few-shot examples,
    and queries the OpenAI API to generate synthetic queries at various temperatures.

    Args:
        document (str): The document text based on which the query is to be generated.
        system_prompt (str): The prompt that guides the AI on how to respond.
        few_shot_examples (str): Preceding examples to prime the model for consistent responses.
        temperatures (list): A list of temperature settings to use for generating diverse outputs.
        length_of_fewshot_prompt (int): The number of few-shot examples provided.

    Returns:
        list: A list of synthetic documents generated at different temperatures.
    """

    time.sleep(1)

    # Initialize a list to store the generated documents
    synth_documents_generated = []

    # Iterate over each temperature value provided in the temperatures list
    for temp in temperatures:
        try:
            # Construct the user prompt by appending the document to the few-shot examples
            user_prompt = few_shot_examples
            user_prompt += "Document: " + document + "\n"
            user_prompt += "Question: "

            # Define the message format for the API request
            messages = [
                {
                    "role": "system",
                    "content": system_prompt
                },
                {
                    "role": "user",
                    "content": user_prompt
                }
            ]

            # Make a request to the OpenAI API with the specified model and temperature
            response = openai.ChatCompletion.create(
                model="gpt-3.5-turbo-16k",  # Consider using "gpt-4" for potentially better results
                messages=messages,
                temperature=temp
            )
            
            # Extract the content from the response and add it to the list of generated documents
            final_response = response["choices"][0]["message"]["content"]
            synth_documents_generated.append(final_response)
        except Exception as e:
            # Handle exceptions by logging the error and pausing before retrying
            print("Error with OpenAI! Waiting 30 seconds...")
            print("Error: " + str(e))
            time.sleep(30)

    # Ensure that at least one document has been generated
    assert len(synth_documents_generated) >= 1, "No documents were generated."

    # Return the list of generated synthetic documents
    return synth_documents_generated

def generate_answer_from_context(document: str, synth_question: str) -> str:
    """
    Generates an answer from a given document context and a synthetic question using the OpenAI API.

    This function attempts to generate an answer up to five times if an error occurs during the API request.
    It introduces a delay before making the request to avoid rapid requests that might lead to API rate limits.

    Args:
    document (str): The document context from which the answer needs to be generated.
    synth_question (str): The synthetic question based on the document context.

    Returns:
    str: The generated answer from the API or an error message if unable to generate after retries.

    Raises:
    Exception: Outputs an error message if the API request fails continuously.
    """

    # Introduce a slight delay before making the API request
    time.sleep(1)

    # Attempt to generate an answer up to 5 times in case of failures
    for attempt in range(5):
        try:
            # Define the system prompt that guides the model's response
            system_prompt = (
                "You are a helpful assistant built by Databricks, you are not human, you are good at helping to answer a query based on the context step by step, the context is a document. "
                "If the query doesn't form a complete question, just say I don't know. "
                "If there is a good answer from the context, try to summarize the context as the answer. "
                "If you don't know the answer, just say I don't know. "
                "If there is no enough information to determine the answer, just say I don't know. "
                "If the context is irrelevant to the question, just say I don't know."
            )
            # Construct the user prompt that includes the synthetic question and the document context
            user_prompt = f"Here is the question: {synth_question}\nHere is the context: {document}?"

            # Prepare the message payload for the API request
            messages = [
                {"role": "system", "content": system_prompt},
                {"role": "user", "content": user_prompt}
            ]

            # Make the API request to OpenAI
            response = openai.ChatCompletion.create(
                model="gpt-3.5-turbo-16k",
                messages=messages,
                temperature=0.0
            )

            # Extract the response content from the API response
            final_response = response["choices"][0]["message"]["content"]
            return final_response

        except Exception as e:
            # Log the error and attempt again if less than 5 attempts have been made
            print(f"Error querying OpenAI on attempt {attempt + 1}: {str(e)}. Attempting again...")

    # If all attempts fail, raise an exception
    raise Exception("Failed to generate an answer from the context after multiple attempts.")

def generate_contradictory_answer_from_context(document: str, synth_question: str) -> str:
    """
    Generates an answer that intentionally contradicts the information provided in the document.

    This function attempts to generate a contradictory answer by instructing the model to create
    false information that disagrees with the content of the document. It tries up to 5 times to get
    a valid response from the OpenAI API.

    Args:
        document (str): The document context from which the answer needs to contradict.
        synth_question (str): The synthetic question based on the document.

    Returns:
        str: The generated contradictory answer from the model or raises an exception after retries.

    Raises:
        Exception: If the API fails to generate an answer after multiple attempts.
    """

    # Introduce a slight delay before making the API request to avoid rapid repeated requests
    time.sleep(1)

    # Attempt to generate an answer up to 5 times in case of failures
    for attempt in range(5):
        try:
            # Construct the system prompt that instructs the model to generate a contradictory answer
            system_prompt = (
                "Create an answer for the given question that contradicts the provided document. "
                "You should create false information that disagrees with what exists within the content of the document."
            )
            # Construct the user prompt that includes the synthetic question and the document context
            user_prompt = f"Question: {synth_question}\nDocument: {document}"
            # Combine system and user prompts
            combined_prompt = f"{system_prompt}\n\n{user_prompt}"

            # Prepare the message payload for the API request
            messages = [
                {"role": "user", "content": combined_prompt}
            ]

            # Make the API request to OpenAI
            response = openai.ChatCompletion.create(
                model="gpt-3.5-turbo-16k",
                messages=messages
            )

            # Extract the response content from the API response
            final_response = response["choices"][0]["message"]["content"]
            return final_response

        except Exception as e:
            # Log the error and attempt again if less than 5 attempts have been made
            print(f"Error querying OpenAI on attempt {attempt + 1}: {str(e)}. Attempting again...")

    # If all attempts fail, raise an exception
    raise Exception("Failed to generate a contradictory answer from the context after multiple attempts.")

def check_generated_answer(answer: str) -> str:
    """
    Checks if the provided answer contains any problematic phrases that indicate uncertainty or lack of knowledge.

    This function iterates through a list of predefined problematic phrases. If any of these phrases are found
    in the answer, it returns "No", indicating the presence of such phrases. If none of the phrases are found,
    it returns "Yes", indicating the answer does not contain problematic content.

    Args:
        answer (str): The answer text to be checked for problematic phrases.

    Returns:
        str: "No" if any problematic phrases are found in the answer, otherwise "Yes".
    """
    # List of phrases that are considered problematic because they indicate uncertainty
    problematic_phrases = ["I don't know", "don't know", "i don't know"]
    
    # Check each phrase in the list to see if it is present in the answer
    for phrase in problematic_phrases:
        if phrase in answer.lower():  # Convert answer to lowercase to ensure the check is case-insensitive
            return "No"
    
    # If no problematic phrases are found, return "Yes"
    return "Yes"
    
def generate_contradictory_answer_examples(queries_dataset: pd.DataFrame, number_of_contradictory_answers_to_generate: int, 
few_shot_examples_for_contradictory_answers=None, api_model=False, synthetic_contradictory_answer_prompt=None, device=None, tokenizer=None, 
model=None, for_fever_dataset=None, for_wow_dataset=None) -> pd.DataFrame:
    """
    Generates a specified number of contradictory answers from a given dataset of queries.

    This function processes a dataset to generate contradictory answers using either a direct approach or a language model-based approach, depending on the availability of few-shot examples. It also removes problematic phrases from the generated answers to ensure the quality of contradictions. The function then merges the original dataset with the newly generated contradictory answers and shuffles the dataset to randomize the order of entries.

    Args:
        queries_dataset (pd.DataFrame): The dataset containing queries for which contradictory answers are to be generated.
        number_of_contradictory_answers_to_generate (int): The number of contradictory answers to generate.
        few_shot_examples_for_contradictory_answers (Optional): Predefined examples to guide the generation process if using a language model.
        device: The device on which the model is loaded (e.g., 'cuda:0').
        tokenizer: The tokenizer used for processing text for the model.
        model: The model used for generating answers.
        for_fever_dataset (Optional): Flag to indicate if the generation is for the FEVER dataset.
        for_wow_dataset (Optional): Flag to indicate if the generation is for the WoW dataset.

    Returns:
        pd.DataFrame: The updated dataset with generated contradictory answers, labels, and shuffled entries.
    """

    def remove_problematic_contradictory_phrases(text: str) -> str:
        """
        Removes predefined problematic phrases from the generated text.

        Args:
            text (str): The text from which problematic phrases are to be removed.

        Returns:
            str: The cleaned text after removing problematic phrases.
        """
        if text is None:
            return text

        problematic_phrases = ["Contradictory Answer:", "The false information created is:", "Incorrect Answer: "]
        text_split = text.split(":")
        if len(text_split) > 1:
            return text_split[1].strip()
        else:
            return text

    # Prepare the dataset for processing
    queries_dataset_copy = queries_dataset.copy()
    queries_dataset_copy = queries_dataset_copy.drop_duplicates(subset=['synthetic_query'])

    # Limit the number of generations to the minimum of the requested number or the dataset size
    number_of_contradictory_answers_to_generate = min(number_of_contradictory_answers_to_generate, len(queries_dataset_copy))
    queries_dataset_copy = queries_dataset_copy.sample(n=number_of_contradictory_answers_to_generate, random_state=42)

    contradictory_answers = []
    contradictory_labels = []

    # Generate contradictory answers for each query in the dataset
    for i in tqdm(range(len(queries_dataset_copy))):
        if api_model:
            contradictory_answer_generated = generate_synthetic_contradictory_answers_api_approach(queries_dataset_copy.iloc[i]['document'], queries_dataset_copy.iloc[i]['synthetic_query'], synthetic_contradictory_answer_prompt, few_shot_examples_for_contradictory_answers, model, for_fever_dataset=for_fever_dataset, for_wow_dataset=for_wow_dataset)
        else:
            contradictory_answer_generated = generate_contradictory_answer_llm_approach(queries_dataset_copy.iloc[i]['document'], queries_dataset_copy.iloc[i]['synthetic_query'], few_shot_examples_for_contradictory_answers, device, tokenizer, model, for_fever_dataset=for_fever_dataset, for_wow_dataset=for_wow_dataset)

        contradictory_answer_generated = remove_problematic_contradictory_phrases(contradictory_answer_generated)

        contradictory_answers.append(contradictory_answer_generated)   
        contradictory_labels.append("No")

    queries_dataset_copy['generated_answer'] = contradictory_answers
    queries_dataset_copy['Answer_Faithfulness_Label'] = contradictory_labels
    queries_dataset_copy['Answer_Relevance_Label'] = contradictory_labels

    # Additional Generation Method: Answer Randomization
    queries_dataset_copy_2 = queries_dataset.copy()
    queries_dataset_copy_2 = queries_dataset_copy_2.drop_duplicates(subset=['synthetic_query'])
    queries_dataset_copy_2 = queries_dataset_copy_2.sample(n=number_of_contradictory_answers_to_generate, random_state=42)
    
    total_answers = queries_dataset[queries_dataset['Answer_Relevance_Label'] == "Yes"]['generated_answer'].tolist()
    total_answers = [answer for answer in total_answers if isinstance(answer, str)]
    total_answers = [str(answer) for answer in total_answers]
    total_answers = [answer for answer in total_answers if len(answer) > 5]

    contradictory_answers_2 = []
    contradictory_labels_2 = []
    for i in tqdm(range(len(queries_dataset_copy_2))):

        random_answer = random.choice(total_answers)
        contradictory_answers_2.append(random_answer)   
        contradictory_labels_2.append("No")

    queries_dataset_copy_2['generated_answer'] = contradictory_answers_2
    queries_dataset_copy_2['Answer_Relevance_Label'] = contradictory_labels_2
    queries_dataset_copy_2['Answer_Faithfulness_Label'] = contradictory_labels_2

    # Combine the original dataset with the two copied datasets
    queries_dataset = pd.concat([queries_dataset, queries_dataset_copy, queries_dataset_copy_2], axis=0, ignore_index=True)

    # Shuffle the combined dataframe to randomize the order of entries
    queries_dataset = queries_dataset.sample(frac=1, random_state=42).reset_index(drop=True)

    return queries_dataset

def generate_contradictory_answer_llm_approach(document: str, question: str, prompt: str, device: torch.device, 
tokenizer: AutoTokenizer, model: AutoModelForCausalLM, for_fever_dataset: bool = False, for_wow_dataset: bool = False) -> str:
    """
    Generates a contradictory answer using a language model approach based on the provided document and question.

    This function constructs a prompt dynamically based on whether it is being used for the FEVER or WoW dataset,
    or a general dataset. It then checks if the total token length exceeds the model's maximum input size and truncates
    the document if necessary. Finally, it generates a contradictory answer using the model.

    Args:
        document (str): The document text based on which the contradictory answer is to be generated.
        question (str): The question text based on the document.
        prompt (str): The initial prompt text to which the document and question will be appended.
        device: The device (CPU/GPU) on which the tokenizer and model are loaded.
        tokenizer: The tokenizer used for encoding the text.
        model: The model used for generating the answer.
        for_fever_dataset (bool, optional): Flag to indicate if the function is being used for the FEVER dataset. Defaults to False.
        for_wow_dataset (bool, optional): Flag to indicate if the function is being used for the WoW dataset. Defaults to False.

    Returns:
        str: The generated contradictory answer text.
    """

    # Construct the initial part of the prompt without the document based on the dataset type
    prompt_without_document = prompt + "Example " + str(prompt.count("Example") + 1) + ":\n"
    if for_fever_dataset:
        prompt_without_document += "Document: \nStatement: \nIncorrect Answer: "
    elif for_wow_dataset:
        prompt_without_document += "Document: \nDialogue: \nIncorrect Response: "
    else:
        prompt_without_document += "Document: \nQuestion: \nIncorrect Answer: "

    # Calculate the token lengths for the prompt, document, and question
    prompt_tokens_length = tokenizer.encode(prompt_without_document, return_tensors='pt').to(device).shape[1]
    document_length = tokenizer.encode(document, return_tensors='pt').to(device).shape[1]
    question_length = tokenizer.encode(question, return_tensors='pt').to(device).shape[1]

    # Check if the total length exceeds the model's maximum input size and truncate if necessary
    if prompt_tokens_length + document_length + question_length + 100 >= 2048:
        reduction_length = prompt_tokens_length + question_length + 100
        encoded_input = tokenizer(document, max_length=2048 - reduction_length, truncation=True, return_tensors='pt')
        truncated_document = tokenizer.decode(encoded_input['input_ids'][0][:2048 - reduction_length]) 
        document = truncated_document.replace("</s>", "")

    # Append the document and question to the prompt
    prompt += "Example " + str(prompt.count("Example") + 1) + ":\n"
    prompt += "Document: " + document + "\n"
    if for_fever_dataset:
        prompt += "Statement: " + question + "\n"
        prompt += "Incorrect Answer: " 
    elif for_wow_dataset:
        prompt += "Dialogue: " + question + "\n"
        prompt += "Incorrect Response: " 
    else:
        prompt += "Question: " + question + "\n"
        prompt += "Incorrect Answer: " 

    # Encode the complete prompt
    input_ids = tokenizer.encode(prompt, max_length=2048, truncation=True, return_tensors='pt').to(device)

    # Check for encoding issues and generate the answer
    if input_ids.shape[0] != 1 or input_ids.shape[1] >= 2048:
        print("Length of problematic input ids: " + str(input_ids.shape))
        print("Length of problematic document: " + str(len(encoded_input['input_ids'][0])))
        assert False

    outputs = model.generate(
        input_ids=input_ids,
        max_length=256,
        do_sample=True,
        top_p=1.0,
        num_return_sequences=1)

    query = tokenizer.decode(outputs[0], skip_special_tokens=True)

    return query


import pandas as pd
import os

def filter_tsv_by_label(tsv_file, label, output_file=None):
    print(f"Reading the TSV file: {tsv_file}")
    # Read the TSV file into a DataFrame
    df = pd.read_csv(tsv_file, sep='\t')
    
    print(f"Filtering out rows where the label '{label}' has empty values")
    # Filter out rows where the specified label has empty values
    initial_row_count = len(df)
    filtered_df = df[df[label].notna()]
    final_row_count = len(filtered_df)
    rows_removed = initial_row_count - final_row_count
    print(f"Number of rows removed: {rows_removed}")
    print(f"Number of rows left: {final_row_count}")
    
    print(f"Removing the label column: {label}")
    # Remove the label column
    filtered_df = filtered_df.drop(columns=[label])
    
    # Remove other label columns
    other_labels = ["Context_Relevance_Label", "Answer_Relevance_Label", "Answer_Faithfulness_Label"]
    other_labels.remove(label)
    removed_labels = [col for col in other_labels if col in filtered_df.columns]
    print(f"Other label columns being removed: {removed_labels}")
    filtered_df = filtered_df.drop(columns=removed_labels)
    
    # Determine the output file path
    if output_file is None:
        base, ext = os.path.splitext(os.path.basename(tsv_file))
        output_file = f"{base}_filtered_{label}{ext}"
    
    print(f"Saving the filtered DataFrame to: {output_file}")
    # Save the filtered DataFrame to a new TSV file in the current directory
    filtered_df.to_csv(output_file, sep='\t', index=False)
    print("Filtering and saving process completed.")



from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import begin
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import filter_dataset
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import preprocess_data
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import load_api_model 
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import load_tokenizer_and_model
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import evaluate_model
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import post_process_predictions
from ares.RAG_Automatic_Evaluation.LLMJudge_RAG_Compared_Scoring import evaluate_and_scoring_data
import torch

machine_label_system_prompt = (
    "Given the following question and document, you must analyze the provided document "
    "and determine whether it is sufficient for answering the question. In your evaluation, "
    "you should consider the content of the document and whether it contains the answer to "
    "the provided question. Output your final verdict by strictly following this format: "
    "'[[Yes]]' if the document is sufficient and '[[No]]' if the document provided is not sufficient."
)

def rag_scoring_config(alpha, num_trials, evaluation_datasets, few_shot_examples_filepath, checkpoints, labels,
model_choice, llm_judge, assigned_batch_size, number_of_labels, gold_label_paths, rag_type, vllm, host_url, request_delay, debug_mode, 
machine_label_llm_model, gold_machine_label_path, prediction_filepaths):
    """
    Configures and runs the RAG scoring process.

    Parameters:
    - alpha: The alpha value for the scoring process.
    - num_trials: Number of trials to run.
    - evaluation_datasets: List of datasets to evaluate.
    - few_shot_examples_filepath: Filepath for few-shot examples.
    - checkpoints: List of model checkpoints.
    - labels: List of labels.
    - model_choice: Choice of model.
    - llm_judge: LLM judge to use.
    - assigned_batch_size: Batch size to use.
    - number_of_labels: Number of labels.
    - gold_label_paths: List of paths to the gold labels.
    - rag_type: Type of RAG.
    - vllm: VLLM to use.
    - host_url: Host URL.
    - request_delay: Delay between requests.
    - debug_mode: Whether to run in debug mode.
    - machine_label_llm_model: Machine label LLM model.
    - gold_machine_label_path: Path to the gold machine labels.
    - prediction_filepaths: List of file paths to save predictions.
    """
    
    if few_shot_examples_filepath == "None" and (llm_judge != "None" or machine_label_llm_model != "None"):
        raise ValueError("'few_shot_examples_filepath' cannot be None if generating machine labels.")
    
    # Validate if either gold_label_paths or gold_machine_label_path is provided
    if gold_label_paths == ["None"] and gold_machine_label_path == "None":
        raise ValueError("Either 'gold_label_paths' or 'gold_machine_label_path' must be provided.")

    # Validate inputs and determine model loading strategy
    if checkpoints:
        if llm_judge and llm_judge != "None":
            print("Warning: Both checkpoint and llm_judge were provided. Using checkpoints.")
        model_loader = lambda chk: load_tokenizer_and_model(model_choice, number_of_labels, chk)
    elif llm_judge and llm_judge != "None":
        model_loader = lambda _: load_api_model(llm_judge, vllm)
    else:
        raise ValueError("No valid model or checkpoint provided.")

    # Use zip only if checkpoints are not empty, otherwise assume only llm_judge is used
    if checkpoints:
        # Here we assume that the length of checkpoints and labels is the same
        pairings = zip(checkpoints, labels)
    else:
        # If no checkpoints, create dummy pairs for labels with None for checkpoint
        pairings = ((None, label) for label in labels)

    all_evaluation_results = []
    
    for idx, (checkpoint, label_column) in enumerate(pairings):

        chekpoint_results = []

        LLM_judge_ratio_predictions = []
        validation_set_lengths = []
        validation_set_ratios = []
        ppi_confidence_intervals = []
        accuracy_scores = []
        for test_set_idx, test_set_selection in enumerate(evaluation_datasets):

            few_shot_examples = begin(evaluation_datasets, checkpoints, labels, few_shot_examples_filepath)

            context_relevance_system_prompt, answer_faithfulness_system_prompt, answer_relevance_system_prompt = filter_dataset(rag_type)
            
            test_set, text_column = preprocess_data(test_set_selection, label_column, labels)

            loaded_model = model_loader(checkpoint)
            if isinstance(loaded_model, tuple):
                model, tokenizer, device = loaded_model
            else:
                model = loaded_model
                tokenizer = None
                device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
            
            eval_model_settings = {
                "test_set": test_set,
                "label_column": label_column,
                "text_column": text_column,
                "device": device,
                "checkpoint": checkpoint,
                "tokenizer": tokenizer,
                "model": model,
                "assigned_batch_size": assigned_batch_size,
                "model_choice": model_choice,
                "context_relevance_system_prompt": context_relevance_system_prompt,
                "answer_faithfulness_system_prompt": answer_faithfulness_system_prompt,
                "answer_relevance_system_prompt": answer_relevance_system_prompt,
                "few_shot_examples_filepath": few_shot_examples_filepath,
                "llm_judge": llm_judge,
                "vllm": vllm,
                "host_url": host_url,
                "request_delay": request_delay,
                "debug_mode": debug_mode
            }

            total_predictions, total_references, results, metric = evaluate_model(eval_model_settings)
            
            post_process_settings = {
                "checkpoint": checkpoint,
                "test_set": test_set,
                "label_column": label_column,
                "total_predictions": total_predictions,
                "labels": labels,
                "gold_label_path": gold_label_paths[test_set_idx] if test_set_idx < len(gold_label_paths) else gold_label_paths[-1],
                "tokenizer": tokenizer,
                "assigned_batch_size": assigned_batch_size,
                "device": device,
                "gold_machine_label_path": gold_machine_label_path,
                "machine_label_system_prompt": machine_label_system_prompt,
                "machine_label_llm_model": machine_label_llm_model,
                "vllm": vllm,
                "host_url": host_url,
                "debug_mode": debug_mode,
                "request_delay": request_delay,
                "few_shot_examples": few_shot_examples
            }

            test_set, Y_labeled_dataset, Y_labeled_dataloader, Y_labeled_predictions, Yhat_unlabeled_dataset, prediction_column = post_process_predictions(post_process_settings) 
            
            evaluate_scoring_settings = {
                "test_set": test_set,
                "Y_labeled_predictions": Y_labeled_predictions,
                "Y_labeled_dataset": Y_labeled_dataset,
                "Y_labeled_dataloader": Y_labeled_dataloader,
                "Yhat_unlabeled_dataset": Yhat_unlabeled_dataset,
                "alpha": alpha,
                "num_trials": num_trials,
                "model": model,
                "device": device,
                "model_choice": model_choice,
                "context_relevance_system_prompt": context_relevance_system_prompt,
                "answer_faithfulness_system_prompt": answer_faithfulness_system_prompt,
                "answer_relevance_system_prompt": answer_relevance_system_prompt,
                "few_shot_examples": few_shot_examples,
                "metric": metric,
                "prediction_column": prediction_column,
                "label_column": label_column,
                "test_set_selection": test_set_selection,
                "LLM_judge_ratio_predictions": LLM_judge_ratio_predictions,
                "validation_set_lengths": validation_set_lengths,
                "validation_set_ratios": validation_set_ratios,
                "ppi_confidence_intervals": ppi_confidence_intervals,
                "accuracy_scores": accuracy_scores,
                "results": results,
                "checkpoint": checkpoint,
                "llm_judge": llm_judge,
                "vllm": vllm,
                "host_url": host_url,
                "request_delay": request_delay,
                "debug_mode": debug_mode,
                "prediction_filepath": prediction_filepaths[test_set_idx] if test_set_idx < len(prediction_filepaths) else prediction_filepaths[-1]
            }
            dataset_results = evaluate_and_scoring_data(evaluate_scoring_settings)
            chekpoint_results.append(dataset_results)
            
        all_evaluation_results.append(chekpoint_results)
            
    return all_evaluation_results


from together import Together
import os

def generate_synthetic_query_api_approach(document: str, synthetic_query_prompt: str, prompt: str, length_of_fewshot_prompt: int, 
                                          model_name: str, percentiles: list, 
                                          for_fever_dataset=False, for_wow_dataset=False,
                                          num_queries: int = 3) -> list:
    """
    Generates synthetic queries based on a document using an API model.

    This function constructs a prompt with the document and generates queries based on the specified dataset type.
    It handles token length limitations by truncating the document if necessary.

    Args:
        document (str): The document to base the query on.
        prompt (str): The initial prompt to the language model.
        length_of_fewshot_prompt (int): The number of few-shot examples already included in the prompt.
        model_name (str): The name of the API model to be used.
        percentiles (list): A list of percentiles for sampling during generation.
        for_fever_dataset (bool, optional): Flag to indicate if the function is used for the FEVER dataset. Defaults to False.
        for_wow_dataset (bool, optional): Flag to indicate if the function is used for the WoW dataset. Defaults to False.
        num_queries (int): Number of synthetic queries to generate per document.

    Returns:
        list: A list of synthetic queries generated by the model.
    """

    synthetic_queries = []

    # Construct the prompt without the document based on the dataset type
    prompt_without_document = prompt + "Example " + str(length_of_fewshot_prompt + 1) + ":\n"
    prompt_without_document += "Document:"

    # Calculate the length of tokens for the prompt and document
    max_tokens = 4096
    token_approx_factor = 4
    prompt_tokens_length = len(prompt_without_document) // token_approx_factor
    document_length = len(document) // token_approx_factor

    # Check if the total length exceeds the model's maximum input size and truncate if necessary
    if prompt_tokens_length + document_length + 100 >= max_tokens:
        truncated_document = document[:(max_tokens - prompt_tokens_length - 100) * token_approx_factor]
        prompt_with_document = prompt_without_document + truncated_document
    else:
        prompt_with_document = prompt_without_document + document

    # Add the final part of the prompt based on the dataset
    if for_fever_dataset:
        prompt_with_document += "\nStatement: "
    elif for_wow_dataset:
        prompt_with_document += "\nDialogue: "
    else:
        prompt_with_document += "\nQuestion: "

    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    client = Together(api_key=TOGETHER_API_KEY)
    
    # Generate queries for each percentile, repeating to reach num_queries
    for _ in range(num_queries):
        for percentile in percentiles:
            success = False
            for attempt in range(5):
                try:
                    chat_completion = client.chat.completions.create(
                        messages=[
                        {"role": "system", "content": synthetic_query_prompt},
                        {"role": "user", "content": prompt_with_document},
                    ],
                    model=model_name,
                    temperature=percentile,
                    stream=False
                )
                    final_response = chat_completion.choices[0].message.content

                    synthetic_queries.append(final_response)
                    success = True
                    break
                except Exception as e:
                    print(f"Error generating synthetic queries: {e}")

            if not success:
                print(f"Failed to generate synthetic queries after 5 attempts for percentile {percentile}")

    return synthetic_queries

def generate_synthetic_answer_api_approach(document: str, question: str, synthetic_answer_prompt: str, prompt: str, 
                                           length_of_fewshot_prompt: int, model_name: str, for_fever_dataset=False, 
                                           for_wow_dataset=False): 
    # Construct the prompt without the document based on the dataset type
    prompt_without_document = prompt + "Example " + str(length_of_fewshot_prompt + 1) + ":\n"
    if for_fever_dataset:
        prompt_without_document += "Document: \nStatement: \nAnswer: "
    elif for_wow_dataset:
        prompt_without_document += "Document: \nDialogue: \nResponse: "
    else:
        prompt_without_document += "Document: \nQuestion: \nAnswer: "

    prompt_tokens_length = len(prompt_without_document)
    document_length = len(document)
    question_length = len(question)

    if prompt_tokens_length + document_length + question_length + 100 >= 4096:
        truncated_document = document[:4096 - prompt_tokens_length - question_length - 100]
        prompt = prompt_without_document + truncated_document
    else: 
        prompt = prompt_without_document + document 
    
    prompt += "Example" + str(length_of_fewshot_prompt + 1) + ":\n"
    prompt += "Document: " + document + "\n"
    if for_fever_dataset:
        prompt += "Statement: " + question + "\n"
        prompt += "Answer: "
    elif for_wow_dataset:
        prompt += "Dialogue: " + question + "\n"
        prompt += "Response: "
    else:
        prompt += "Question: " + question + "\n"
        prompt += "Answer: "

    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    client = Together(api_key=TOGETHER_API_KEY)
    
    for attempt in range(5):
        try:
            chat_completion = client.chat.completions.create(
                messages=[
                {"role": "system", "content": synthetic_answer_prompt},
                {"role": "user", "content": prompt},
            ],
            model=model_name,
            stream=False
        )
            # responses = []
            
            # for chunk in chat_completion:
            #     responses.append(chunk.choices[0].delta.content)

            # final_response = " ".join(responses)

            final_response = chat_completion.choices[0].message.content
            
            return final_response
        except Exception as e:
            print(f"Error generating synthetic queries: {e}")
            continue

def generate_synthetic_contradictory_answers_api_approach(document: str, question: str, synthetic_contradictory_answer_prompt: str, fewshot_examples: str, 
                                           model_name: str, for_fever_dataset=False, for_wow_dataset=False) -> str:
    """
    Generates a synthetic contradictory answer using an API-based approach based on the provided document and question.

    This function constructs a prompt dynamically based on whether it is being used for the FEVER or WoW dataset,
    or a general dataset. It then sends the prompt to the API to generate a contradictory answer.

    Args:
        document (str): The document text based on which the contradictory answer is to be generated.
        question (str): The question text based on the document.
        synthetic_answer_prompt (str): The initial prompt text to which the document and question will be appended.
        fewshot_examples (str): Few-shot examples to include in the prompt for the API.
        api_url (str): The API endpoint URL.
        api_key (str): The API key for authentication.
        model_name (str): The model name to be used in the API.
        for_fever_dataset (bool, optional): Flag to indicate if the function is being used for the FEVER dataset. Defaults to False.
        for_wow_dataset (bool, optional): Flag to indicate if the function is being used for the WoW dataset. Defaults to False.

    Returns:
        str: The generated contradictory answer text.
    """

    prompt = synthetic_contradictory_answer_prompt + fewshot_examples
    prompt += "Example " + str(prompt.count("Example") + 1) + ":\n"
    if for_fever_dataset:
        prompt += "Document: \nStatement: \nIncorrect Answer: "
    elif for_wow_dataset:
        prompt += "Document: \nDialogue: \nIncorrect Response: "
    else:
        prompt += "Document: \nQuestion: \nIncorrect Answer: "

    prompt_token_length = len(prompt)
    document_token_length = len(document)
    question_token_length = len(question)

    if prompt_token_length + document_token_length + question_token_length + 100 >= 4096:
        truncated_document = document[:4096 - prompt_token_length - question_token_length - 100]
        prompt += truncated_document
    else: 
        prompt += document
    
    prompt += "Example " + str(prompt.count("Example") + 1) + ":\n"
    prompt += "Document: " + document + "\n"
    if for_fever_dataset:
        prompt += "Statement: " + question + "\n"
        prompt += "Incorrect Answer: "
    elif for_wow_dataset:
        prompt += "Dialogue: " + question + "\n"
        prompt += "Incorrect Response: "
    else:
        prompt += "Question: " + question + "\n"
        prompt += "Incorrect Answer: "
    
    TOGETHER_API_KEY = os.getenv("TOGETHER_API_KEY")

    client = Together(api_key=TOGETHER_API_KEY)
    
    for attempt in range(5):
        try:
            chat_completion = client.chat.completions.create(
                messages=[
                {"role": "system", "content": synthetic_contradictory_answer_prompt},
                {"role": "user", "content": prompt},
            ],
            model=model_name,
            stream=False
        )
            # responses = []
            
            # for chunk in chat_completion:
            #     responses.append(chunk.choices[0].delta.content)

            # final_response = " ".join(responses)

            final_response = chat_completion.choices[0].message.content

            return final_response
        except Exception as e:
            print(f"Error generating synthetic contradictory answer: {e}")
            continue